% Anisotropic Fitted Local Likelihood (FLL) Denoising Demo (demo_FittedLocalLikelihoodDenoising.m)
%
% V. Katkovnik - Tampere University of Technology - July 2006
% -----------------------------------------------------------------------
%
% Performs the anisotropic FLL denoising on observations subject to
% Gaussian, Poisson and Binary Bernoilli distributions
% For color imaging the colors are processed independently
% Papers: Vladimir Katkovnik,  Vladimir Spokoiny "SPATIALLY ADAPTIVE NON-GAUSSIAN IMAGING VIA FITTED LOCAL
% LIKELIHOOD TECHNIQUE", Proceedings of SMMSP 2006, Florence.
% Vladimir Katkovnik,  Vladimir Spokoiny, "SPATIALLY ADAPTIVE NON-GAUSSIAN IMAGING VIA FITTED LOCAL
% LIKELIHOOD TECHNIQUE", IEEE, Trans Signal Processing, 2006, submitted.
%
% Observation model:
% z : noisy observation
% y : true image (unknown)
% n : Gaussian white noise
%
% Other key variables:
% y_hat   : anisotropic "fused" estimate
% y_hat_Q : adaptive directional estimate
% h_opt_Q : adaptive directional scales



clear all
close all
tic
directional_resolution=8;  % number of directions used in the anisotropic estimation
fusing=1;                  % fusing type   (1 classical fusing (inverse variance weights), 2 piecewise regular)
filtering=1;               % swith on 1 switch off 0; median filtering of the adaptive window sizes
kappa=1; % do not change this parameter 
%%%%%%%%%% Selection of distribution %%%%%%%%%%%%%%%%%%%%%%
% Types of distributions , 1 only for one from these distributions
poisson=1;
binary=0;
gaussian=0;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%% DATA DISTRIBUTIONS %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if poisson==1
   pdf=2;% for Poissonian distribution in observation modeling
   kappa=255/10; % scalling of Poissonian noise
  
   h1=floor((1.5).^[1:7]); % Geometrical scale sequence: default 
   h2=max(1,ceil(h1*tan(.75*pi/directional_resolution))); % Windows width; For texture images

% Threshold parameters approximating linear decreasing sequence
gamma_max=1.4; gamma_min=.6; %  ISNR=11.35
gamma_max=1.7; gamma_min=.2; % ISNR=11.46, .75

%gamma_max=2; gamma_min=1; % 
delta_gamma=(gamma_max-gamma_min)/(length(h1)-2);
gammaICI=[gamma_max-delta_gamma*[0:(length(h1)-2)]];

gammaICI=[1.6000    1.3696    1.1392    0.9088    0.6784    0.4480]; % selected for r=1; alpha=1.; 

end
if gaussian==1
   pdf=1;
  
   h1=floor((1.5).^[1:7]);
 
   h2=max(1,ceil(h1*tan(.75*pi/directional_resolution)));  % % Windows width; For texture images
   
   sigma_noise=0.1;
  

% threshold parameters, approximate linear decreasing sequence
  
   gamma_max=2.5; gamma_min=1;  % for corrected h2
   delta_gamma=(gamma_max-gamma_min)/(length(h1)-2);
   gammaICI=[ gamma_max-delta_gamma*[0:(length(h1)-2)]];
    
   
   gammaICI=[3.0000    2.6400    2.2800    1.9200    1.5600    1.2000]; % selected for r=1; alpha=1;

end
if binary==1
   pdf=3; % Binary distribution
  
  
   h1=floor((1.5).^[1:7]);


  h2=max(1,ceil(h1*tan(.75*pi/directional_resolution)));  % % Windows width; For texture images
  delta=.85 % noise-level scale parameter of the binary distribution

   gammaICI=[0.7000    0.6860    0.6720    0.6580    0.6440    0.6300];  % thresholds corresponding to r=1/2, alpha=1


end

%--------------------------------------------------------------------------
% WINDOWS PARAMETERS
%--------------------------------------------------------------------------
lenh=length(h1);
sig_winds=[1.5*ones(size(h1)); 1.5*ones(size(h2))];    % Gaussian window parameter
beta=1;                     % Parameter of window 6

window_type=112;  % window=1 for uniform, window=2 for Gaussian

% window=6;  %for exponentions with beta
% window=8 for Interpolation
% window=11 for sectorial windows
%window_type=112
%window=112; % for sectorial uniform windows


TYPE=10;            % TYPE IS A SYMMETRY PARAMETER OF THE WINDOW
% 00 SYMMETRIC
% 10 NONSYMMETRIC ON X1 and SYMMETRIC ON X2
% 11 NONSYMMETRIC ON X1,X2  (Quadrants)

%--------------------------------------------------------------------------
% MODELLING
%--------------------------------------------------------------------------
% Test-Images
 yy=im2double(imread('image_Cameraman256.png'));
% yy=ones(size(yy));
% yy=im2double(imread('image_Lena256.png'));

yy=im2double(imread('image_Lena512.png'));
%  yy=im2double(imread('image_Peppers512.png'));
% yy=im2double(imread('IMAGE_boats512.png'));
 % yy=im2double(imread('image_Cheese128.png')); % For Bernoulli distribution
 %yy=.8*im2double(imread('image_Cheese128.png'))+0.2; % For Poissonian and
% 
% yy=im2double(imread('image_Boats512.png'));
% yy=im2double(imread('barbara_port.png'));
% yy=im2double(imread('kodim23.png')); % parrots
% yy=im2double(imread('kodim19.png')); % lighhouse
% yy=im2double(imread('kodim04.png')); % girl

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


disp(' ')
disp(' ')
disp('-----------------------------------------------------------------------------------')
disp(' Anisotropic FLL Denoising Demo ')
disp('-----------------------------------------------------------------------------------')
disp(' ')


if size(yy,3)>1
    nncolor=3;
    yy = yy(1:2:end,1:2:end,:);
else
nncolor=1;    
end

% if nncolor==3; % colorspace transform
% yy=rgb2Oppon(yy);
% end


for s_color=1:nncolor % 1:3 for color images
s_color
y=yy(:,:,s_color);
   
    

%--------------------------------------------------------------------------
% MODELLING
%--------------------------------------------------------------------------

[size_z_1,size_z_2]=size(y);

state = 2055615866;
rand('state', state);
randn('state', state);

if gaussian==1
    n=repmat(sigma_noise,size_z_1,size_z_2).*randn(size(y));
    z = y + n;
end
 


if poisson==1  
z=poissrnd(kappa*y)/kappa;

end

if binary==1 % binary distribution with the parameter  y=0.5*(1-delta)+y*delta;
    z1=unifrnd(zeros(size(y)),ones(size(y)));
    y=0.5*(1-delta)+y*delta;
    II=(z1<y);
    z=zeros(size(y));
    
    z(II==1)=1;
end   
    sigma=function_stdEst2D(z);   %%% estimates noise standard deviation
disp('Initial (noisy) observation criteria values:')

function_Errors(y,z,z,2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%--------------------------------------------------------------------------
% LPA KERNELS 
%--------------------------------------------------------------------------
% calling kernel creation function
[kernels, kernels_higher_order]=function_CreateLPAKernels([0 0],h1,h2,TYPE,window_type,directional_resolution,sig_winds,beta);
disp('  ')
disp('kernels created');

sigmaiter=repmat(sigma,size_z_1,size_z_2);
stop_condition=0;



clear yh h_opt_Q y_hat_Q var_opt_Q stdh

YICI_Final1=0; var_inv=0;         YICI_Final2=0;
CWW=0;
CWW2=0;


htbar=timebar('FLL running','Progress');
figure
figure_number=gcf;
imshow([z,zeros(size(z))])
disp(' ')
disp(' ')
disp('Directional estimates criteria values:')
disp(['  (Computing direction ',num2str(1),' ... )'])
%%%%% loops %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  YICI_Final1=0;  YICI_Final2=0;
for s1=1:directional_resolution     % directional index
          
    for s2=1:lenh     % kernel size index
        gh=kernels_higher_order{s1,s2,1}(:,:,1);   %gets single kernel from the cell array
       
        ghorigin(s1,s2)=gh((end+1)/2,(end+1)/2);
        bound1=min([(find(sum(gh~=0,2)));abs(find(sum(gh~=0,2))-size(gh,1)-1)]); % removes unnecessary zeroes
        bound2=min([(find(sum(gh~=0,1))),abs(find(sum(gh~=0,1))-size(gh,2)-1)]); % removes unnecessary zeroes
        gh=gh(bound1:size(gh,1)-bound1+1,bound2:size(gh,2)-bound2+1);            % removes unnecessary zeroes
        yh(1:size_z_1,1:size_z_2,s2)= conv2(z,gh,'same'); % Estimation
        N_Kulback(s2)=sum(gh(:)>0);
   % N_Kulback(s2)=h1(s2);
       stdh(:,:,s2)=repmat(sigma*(sum(gh(:).^2))^0.5,size_z_1,size_z_2);  % Std of the estimate
    end %% for s2, window sizes
    if filtering==1
[YICI,h_opt,std_opt]=function_FLL_for_ML_WOS(yh,stdh,gammaICI,N_Kulback,sigma,pdf,kappa,2*(s1-1)*pi/directional_resolution);
    else
        [YICI,h_opt,std_opt]=function_FLL_for_ML_WOS(yh,stdh,gammaICI,N_Kulback,sigma,pdf,kappa);
    end

    aaa=reshape(ghorigin(s1,h_opt),size(h_opt));  %origin weight for optimal kernels
    y_hat_Q(:,:,s1)=YICI; 
    h_opt_Q(:,:,s1)=h_opt;   
    var_opt_Q(:,:,s1)=(std_opt.^2+eps);
    YICI_Final1=YICI_Final1+y_hat_Q(:,:,s1)./var_opt_Q(:,:,s1);            %% FUSING %%%%%
    %II=h_opt_Q(:,:,s1)>=2;
    YICI_Final2=YICI_Final2+(y_hat_Q(:,:,s1)./var_opt_Q(:,:,s1)-z.*aaa./var_opt_Q(:,:,s1));            %% FUSING 2 %%%%%
    var_inv=var_inv+1./var_opt_Q(:,:,s1);
    CWW=CWW+(aaa./var_opt_Q(:,:,s1));
    CWW2=CWW2+(aaa./var_opt_Q(:,:,s1));  

    timebar(htbar,s1/directional_resolution);
    if gcf~=figure_number,        figure(figure_number), end
    imshow([YICI,(h_opt-1)/(lenh-1)]);
    title(['Directional adaptive estimates and corresponding adaptive scales (\theta=', num2str(360/directional_resolution*(s1-1)),'^o)'])
    [Err_y_Q,Err_labels]=function_Errors(y,YICI,z);  %% computes error criteria
    %%%% THIS PART HANDLES THE RESULTS TABLE DISPLAY
    if s1<directional_resolution
        Err_Wait=repmat('  wait... ',[size(Err_y_Q,1) 1]);
        size_Err_Wait=size(Err_Wait,2);
        Err_Wait=[repmat(' ',[1 size_Err_Wait]);Err_Wait];
    else
        Err_Wait=repmat([],[size(Err_y_Q,1)+1 1]);
    end
    if s1==1
        number_of_digits=4;  % number of digits for results in the table
        Err_y_Q_str=num2str(Err_y_Q,number_of_digits);
        tab_content=[['Dir.# ',repmat(' ',[1 size(Err_labels,2)-6]);Err_labels],repmat(' ',[size(Err_y_Q,1)+1 1]),[repmat(' ',[1 floor((size(Err_y_Q_str,2))/2)]),num2str(s1),repmat(' ',[1,ceil((size(Err_y_Q_str,2))/2)-1]);Err_y_Q_str],Err_Wait];
    else
        Err_y_Q_str=num2str(Err_y_Q,number_of_digits);
        disp(sprintf(repmat('\b',[1 numel(tab_content)+size(tab_content,1)+32]))) % clears previous table
        tab_content=[tab_content(:,1:end-size_Err_Wait),repmat(' ',[size(Err_y_Q,1)+1 2]),[repmat(' ',[1 1-numel(num2str(s1))+floor((size(Err_y_Q_str,2))/2)]),num2str(s1),repmat(' ',[1,ceil((size(Err_y_Q_str,2))/2)-1]);Err_y_Q_str],Err_Wait];
    end
    if s1==1
        disp(sprintf(repmat('\b',[1 32]))) % clears previous table
    end
    if s1<directional_resolution
        disp(['  (Computing direction ',num2str(s1+1),' ... )'])
    end
    if s1==directional_resolution
        disp('   ')
    end
    disp(tab_content);
    %%%% END OF RESULTS TABLE DISPLAY
end     %for s1 directions
disp(' ')
disp('Filtering completed.         ')
YICI_Final1=YICI_Final1./var_inv;
YICI_Final2=(YICI_Final2+z.*CWW/directional_resolution)./(var_inv-CWW+CWW./directional_resolution);           %% FUSING 2 %%%%%

if fusing==1, y_hat=YICI_Final1; end
if fusing==2, y_hat=YICI_Final2; end
%%%%%%%%%%%%%   END OF ALGORITHM   %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close(figure_number)
close(htbar);
[Err_y_Q,Err_labels]=function_Errors(y,YICI_Final1,z,2); 'YICI_Final1'
[Err_y_Q,Err_labels]=function_Errors(y,YICI_Final2,z,2); 'YICI_Final2'

% disp('  ')
% disp('  ')
% disp('Final anisotropic "fused" estimate criteria values:')
% 
% function_Errors(y,y_hat,z,2);
% disp('  ')

figure
subplot(1,2,1), imshow(h_opt_Q(:,:,1),[]), title(['Adaptive scales, \theta=', num2str(360/directional_resolution*(1-1)),'^o'])
subplot(1,2,2), imshow(y_hat_Q(:,:,1)), title(['Adaptive directional estimate, (\theta=', num2str(360/directional_resolution*(1-1)),'^o)'])

figure
subplot(1,2,1), imshow(h_opt_Q(:,:,4),[]), title(['Adaptive scales, \theta=', num2str(360/directional_resolution*(4-1)),'^o'])
subplot(1,2,2), imshow(y_hat_Q(:,:,4)), title(['Adaptive directional estimate, (\theta=', num2str(360/directional_resolution*(4-1)),'^o)'])
figure
subplot(1,2,1)
imshow(z); title('Noisy observation')
subplot(1,2,2)
imshow(y_hat); title('Anisotropic "fused" estimate')

    y_est_color(:,:,s_color)=y_hat;
    z_color(:,:,s_color)=z;
end
toc

% if nncolor==3; % colorspace inverse
%  y_est_color=Oppon2rgb(y_est_color);
%  yy=Oppon2rgb(yy);
% end

figure, imshow(y_est_color,[]),title('Estimate')
figure, imshow(z_color,[]),title('Noisy')
figure, imshow(yy,[]),title('True')
